/* Compile with "gcc -O -DMODULE -D__KERNEL__ -c qqs.c" */

#include <linux/config.h>

#include <linux/module.h>
#include <linux/version.h>

#include <linux/types.h>
#include <linux/fs.h>
#include <linux/mm.h>    /* for verify_area */
#include <linux/errno.h> /* for -EBUSY */
#include <linux/ioport.h> /* for -EBUSY */
#include <asm/segment.h> /* for put_user_byte */
#define SLOW_IO_BY_JUMPING
#include <asm/io.h>
#define inportb(x) inb_p(x)
#define outportb(x,y) outb_p(y,x)

#define QQS_MAJOR 0
#define QQS_PORT 0x130
#define QQS_IRQ  11

/* port number QQS_PORT + ...:
   0   read: next data byte;
       write: jump to next line;
   1   read: bit 7:   busy;
                 0-6: line number in buffer;
       write: reset line number;
   2   bit 0: teletext on;
           1: tuner on;
           2: SDA - I2C serial data;
           3: SCL - I2C serial clock;
       read bit 4: line number overflow;
                5: MMV line RAM write ?
                6: MMV line clamp ?
                7: composite sync bit ?
       write bit 4: audio/video mute;
                 5: audio mute;
                 6: unused;
                 7: digital out.
 */

int major, ixin, ixout;        /* device major, index for in and out */
unsigned char buffer[128][44],    /* the buffer                         */
         par_ham[256],         /* parity and hamming decoding        */
         chapter, page, sub;   /* current page                       */

const char *qqs_name = "qqs";

#include <time.h>
#include <linux/delay.h>
   

void init_par_ham(void) {
/* initialise table for parity and hamming
   entry & 0x80 -> odd parity
   entry & 0x40 -> invalid hamming code (> 1 bit error)
   entry & 0x0f -> Hamming decoded                       */

   unsigned char mask, i;
/* first define parity bit */

   par_ham[0] = 0x7f;
   for (mask = 1; mask; mask<<=1)
      for (i=0; i<mask; i++)
         par_ham[mask+i] = par_ham[i] ^ 0x80;          /* got this? */

   for (i=0; i < 16; i++) {               /* stole this from Martin */
      unsigned char val
        = ( !(i & 1) ^ !!(i & 4) ^ !!(i & 8)) << 0 | !!(i & 1) << 1 |
          ( !(i & 1) ^ !!(i & 2) ^ !!(i & 8)) << 2 | !!(i & 2) << 3 |
          ( !(i & 1) ^ !!(i & 2) ^ !!(i & 4)) << 4 | !!(i & 4) << 5 |
          (!!(i & 2) ^ !!(i & 4) ^ !!(i & 8)) << 6 | !!(i & 8) << 7;
      par_ham[val] &= (0x80 | i);
      for (mask = 1; mask; mask <<= 1)
         par_ham[val ^ mask] &= (0x80 | i);
   }
}

int read_ham(void) {
/* read Hammed byte from device */    
    unsigned char b0, b1;

    b0 = par_ham[inportb(QQS_PORT)];
    b1 = par_ham[inportb(QQS_PORT)];

/*  b0 and b1 are hammed nibbles, decode or -1 if error */

    if ((b0 | b1) & 0x40)
       return -1;
    else
       return (b1 & 0x0f) << 4 | (b0 & 0x0f);
}

void qqs_intr(int irq, void *dev_id, struct pt_regs *regs) {

/* the interupt handler */

  int numlines, j;
  if (irq != QQS_IRQ) return;
  numlines = inportb(QQS_PORT + 1) & 0x7f;
  outportb(QQS_PORT + 1, 0);          /* reset line number */

  while (numlines > 0) {
      unsigned char c0, c1, c2, br, ag;
      int linchap;
      c0 = inportb(QQS_PORT);         /* video line, not used */
      linchap = read_ham();           /* line & chapter       */
      if (linchap >= 0) {
         unsigned char chap, line, inbuf=4;
         chap = linchap & 0x07;
         line = linchap >> 3;
         if (! line) {                /* line 0 */
            int hours, minutes, contr;
            chapter = chap;
            page = read_ham();
            minutes  = read_ham();
            hours    = read_ham();
            contr    = read_ham();
            
            contr =  contr << 7
                  | (hours & 0xc0) >> 1
                  | (minutes & 0x80) >> 3;  /* so bit 4 = C4 etc. */
            hours   &= 0x3f;
            minutes &= 0x7f;
            sub = minutes;
            buffer[ixin][4] = hours;    /* abuse 4 bytes to store info */
            buffer[ixin][5] = minutes;
            buffer[ixin][6] = contr >> 8;
            buffer[ixin][7] = contr & 0xff;
            for (inbuf = 8; inbuf<12; inbuf++)
                 buffer[ixin][inbuf]=0;
         }
         
         if (chap == chapter) {    /* if not, we don't know what this is */
            buffer[ixin][0] = chap;
            buffer[ixin][1] = page;
            buffer[ixin][2] = sub;
            buffer[ixin][3] = line;
            for (j=inbuf; j<44; j++)
                buffer[ixin][j] = inportb(QQS_PORT);
            j = (ixin + 1) & 127;
            if (j != ixout) ixin = j;
         }
      }
      outportb(QQS_PORT,0);                      /* go to next line */
      numlines--;
  }
  outportb(QQS_PORT + 1, 0);                        /* reset buffer */
}
 
static int read_qqs (struct inode * node, struct file * file,
                     char *buf, int count) {
  int left, avail, wanted;

  if (verify_area(VERIFY_WRITE, buf, count) == -EFAULT) return -EFAULT;
  avail = (ixin - ixout) & 127;            /* lines in buffer */
  wanted= count / 44;                      /* lines wanted    */
  if (wanted > avail) wanted = avail;
  count = wanted * 44;                     /* # bytes to give */
  while (wanted > 0) {
     memcpy_tofs(buf, buffer[ixout], 44);
     ixout = (ixout + 1) & 127;
     wanted--;
     buf+=44;
  }
  return count;
}

static int write_qqs (struct inode * inode, struct file * file,
                      const char * buf, int count) {
  return count;                               /* just junked */
}

#include "tuner.c"

static int lseek_qqs(struct inode * inode, struct file * file,
                     off_t offset, int orig) {

/* This is very weird!!! We seek a station, offset = frequency in kHz */

   if (!SelectChannel( (unsigned long) offset * 1000))
      printk ("Problem selecting channel\n");
   return 0;
}

static int open_qqs (struct inode * node, struct file * file) {
  MOD_INC_USE_COUNT;

  ixin = ixout = 0;

  outportb(QQS_PORT + 2, 1);         /* teletext on  */
  outportb(QQS_PORT + 1, 0);         /* reset buffer */

  return 0;
}

static void release_qqs (struct inode * node, struct file * file) {

  outportb(QQS_PORT + 2, 0);        /* teletext & tuner off */

  MOD_DEC_USE_COUNT;
}

static struct file_operations qqs_fops = {
  lseek_qqs,
  read_qqs,
  write_qqs,
  NULL,
  NULL,
  NULL,
  NULL,
  open_qqs,
  release_qqs,
  NULL,
  NULL,
  NULL,
  NULL
};

int qqs_check(void) {
  int ii;
  /* test presence of ports */
  for (ii = 0; ii < 4; ii++)
     if(inportb(QQS_PORT + ii) != 0xff)
        break;
  if (ii == 4) return(-ENODEV);

  outportb(QQS_PORT + 2, 0);            /* disable teletext */
                            /* test reset/increment ability */
  outportb(QQS_PORT + 1, 0);                       /* reset */
  for (ii = 0; ii < 32; ii++) {
     if((inportb(QQS_PORT + 1) & 0x1f) != ii)
        return(-EIO);
     outportb(QQS_PORT, 0);  /* next */
  }

  outportb(QQS_PORT + 1, 0); /* reset */
  if((inportb(QQS_PORT + 1) & 0x1f) != 0)
      return(-EIO);
  return 0;
}

int init_module(void) {
  int stat;
  if (stat=check_region(QQS_PORT, 4)) {
    printk ("Can't get ports\n");
    return stat;
  }

  request_region(QQS_PORT, 4, qqs_name);

  if (stat=qqs_check()) {
    printk ("No card found\n");
    return stat;
  }

  if (!InitTune()) {
    printk ("Tuner malfunction\n");
    return -EIO;
  }

  init_par_ham();

  if ((major = register_chrdev(QQS_MAJOR, qqs_name, &qqs_fops)) == -EBUSY) {
    printk ("Unable to get major for qqs device\n");
    return -EIO;
  }
  request_irq(QQS_IRQ, qqs_intr, 0, qqs_name, NULL);
  return 0;
}

void cleanup_module(void) {
  free_irq(QQS_IRQ, NULL);
  unregister_chrdev(major, qqs_name);
  release_region(QQS_PORT, 4);
}
