/*
 * xevents.c: Callbacks for XView-objects. The main work of the XView-version is done here.
 *
 * $Id: xevents.c,v 1.6 1999/10/04 23:10:05 mb Exp mb $
 *
 * Copyright (c) 1994-97 Martin Buck  <martin.buck@bigfoot.com>
 * Read COPYING for more information
 *
 */

#include <unistd.h>
#include <ctype.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <sys/time.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <xview/xview.h>
#include <xview/frame.h>
#include <xview/panel.h>
#include <xview/notify.h>
#include <xview/notice.h>
#include "vtx_assert.h"
#include "safe_malloc.h"
#include "misc.h"
#include "cct.h"
#include "tuner.h"
#include "vtxtools.h"
#include "vtxdecode.h"
#include "vtxqueue.h"
#include "cache.h"
#include "search.h"
#include "toptext.h"
#include "fastext.h"
#include "config.h"
#include "dialog.h"
#include "hotlist.h"
#include "spool.h"
#include "fileio.h"
#include "selection.h"
#include "xvtools.h"
#include "xevents.h"
#include "xinit.h"

#include "titlepage.h"


#define CHECK_NAME_CYCLE 5
#define DRAG_THRES 5


vtxpgwin_t vtxwin;
vtxpage_t page;
byte_t vtx_buffer[VTX_PAGESIZE];
int keep_closed = FALSE, do_dither, tuner_dev_open;
XComposeStatus compose_status;

static int update_count, doabort;
static volatile int reload_intv_file = TRUE;
static vtxpage_t curr_disp;


void
xquit(void) {
  xv_destroy_safe(frame);
}


void
xabort(void) {
  doabort = TRUE;
  xv_destroy_safe(frame);
}


int
report_cct_error(int err, const char *err_func_str) {
  Xv_notice notice;
  char *catmsg;
  int retval;
  
  catmsg = sstrdup("Error: Videotext-driver returned\n");
  switch (err) {
    case CCTERR:
      catmsg = sstrapp(catmsg, "I/O-");
    break;
    case CCTEINVAL:
      catmsg = sstrapp(catmsg, "invalid argument ");
    break;
    default:
      catmsg = sstrapp(catmsg, "unknown ");
    break;
  }
  catmsg = sstrapp(catmsg, "error in function ");
  catmsg = sstrapp(catmsg, err_func_str);
  catmsg = sstrapp(catmsg, "\nfor device ");
  catmsg = sstrapp(catmsg, cct_device);
  notice = (Xv_notice)xv_create(frame, NOTICE,
      NOTICE_MESSAGE_STRING, catmsg,
      NOTICE_BUTTON_YES, "Retry",
      NOTICE_BUTTON_NO, "Close device",
      XV_SHOW, TRUE,
      NULL);
  switch ((int)xv_get(notice, NOTICE_STATUS)) {
    case NOTICE_YES:
      retval = FALSE;
    break;
    case NOTICE_NO:
      retval = TRUE;
    break;
    default:
      assert(0);
    break;
  }
  xv_destroy_safe(notice);
  free(catmsg);
  return retval;
}


void
report_tuner_error(int err, const char *err_func_str) {
  Xv_notice notice;
  char *catmsg;
  
  catmsg = sstrdup("Error: Videotext/tuner-driver returned\n");
  switch (err) {
    case TUNERR:
      catmsg = sstrapp(catmsg, "I/O-");
    break;
    case TUNEINVAL:
      catmsg = sstrapp(catmsg, "invalid argument ");
    break;
    default:
      catmsg = sstrapp(catmsg, "unknown ");
    break;
  }
  catmsg = sstrapp(catmsg, "error in function ");
  catmsg = sstrapp(catmsg, err_func_str);
  catmsg = sstrapp(catmsg, "\nfor device ");
  catmsg = sstrapp(catmsg, tuner_device);
  notice = (Xv_notice)xv_create(frame, NOTICE,
      NOTICE_MESSAGE_STRING, catmsg,
      XV_SHOW, TRUE,
      NULL);
  xv_destroy_safe(notice);
  free(catmsg);
}


Notify_value
signal_handler(Notify_client client, int signum, Notify_signal_mode when) {
  /* Ignore SIGPIPE */
  if (signum == SIGUSR1 && intv_fname) {
    reload_intv_file = TRUE;
  }
  return NOTIFY_DONE;
}


void
font_menu_proc(Menu menu, Menu_item menu_item) {
  int font;
  
  font = (int)xv_get(menu_item, MENU_CLIENT_DATA);
  xv_set(frame,
      XV_WIDTH, 41 * vtxfonts[font].width + (intv_fname ? 0 : RPANEL_WIDTH) +
          4 * WIN_DEFAULT_BORDER_WIDTH,
      XV_HEIGHT, 26 * vtxfonts[font].height + CANVAS_OFFSET,
      NULL);
}


void
update_menu_proc(Menu menu, Menu_item menu_item) {
  update_count = 0;
  config_set_update_speed((int)xv_get(menu_item, MENU_CLIENT_DATA));
}


void
checkbox_proc(void) {
  int val;

  val = (int)xv_get(checkbox, PANEL_VALUE);
  if (vtxwin.stopped != (val & 1)) {
    vtxwin.stopped = val & 1;
    if (vtx_dev_open && !vtxwin.stopped) {
      /* Update the display immediately (new pages may have arrived in the cache in the meantime) */
      get_page(PRI_FGROUND_NH, dau_status[0].page, dau_status[0].subpage, 0, vtx_buffer, &vtxwin);
    }
  }
  if (vtxwin.hidden != !(val & 2)) {
    vtxwin.hidden = !(val & 2);
    vtxwin.update = SCR_UPDATE;
  }
  xv_set(intv_reveal_toggle, PANEL_VALUE, !vtxwin.hidden, NULL);
}


void
intv_reveal_proc(void) {
  xv_set(checkbox, PANEL_VALUE, (int)xv_get(checkbox, PANEL_VALUE) ^ 2, NULL);
  checkbox_proc();
}


/* new_station gets called after a new station has been recognized/selected with the new
 * station number in station. full_reset should be set if any station was recognized
 * before the new station.
 */
int
new_station(int station, int full_reset) {
  int old_station;
  
  old_station = hotlist_get_current();
  if (full_reset) {
    write_cache();
    clear_cache = TRUE;
    remove_priority(PRI_ALL);
    cache_reset();
    history_reset();
    search_reset();
    timeout_reset();
    update_pagenumdisp(-1, -1, -1, -1);
  }
  hotlist_set_current(station);
  search_new_station();
  station_selected = FALSE;
  if (full_reset && get_page(PRI_FGROUND, hotlist_get_index_page(), 0, 0, vtx_buffer, &vtxwin) < 0)
    return -1;
  toptext_reset();
  toptext_ok_notify();
  fastext_reset();
  fastext_ok_notify();
  if (vtx_dev_open) {
    CCTCHK(cct_set_virtual(!hotlist_get_fastdisable()), "cct_set_virtual", return -1);
  }
  if (station && !hotlist_get_topdisable() &&
      get_page(PRI_TOP, NUM_BASICTOP, -1, -1, NULL, NULL) < 0)
    return -1;
  hotlist_put_queue();
  return 0;
}


void
reset_proc(void) {
  if (vtx_dev_open) {
    cct_close();
    vtx_dev_open = FALSE;
  }
  new_station(0, TRUE);
}


void
get_index_page(void) {
  get_page(PRI_FGROUND, hotlist_get_index_page(), 0, 0, vtx_buffer, &vtxwin);
}


void
open_device(void) {
  if (!vtx_dev_open) {
    int spool_only;
    
    keep_closed = FALSE;
    spool_only = config_get_spool_only();
    config_set_spool_only(FALSE);
    get_index_page();
    config_set_spool_only(spool_only);
  }
}


void
close_device(void) {
  keep_closed = TRUE;
  cct_close();
  vtx_dev_open = FALSE;
  tuner_close();
  tuner_dev_open = FALSE;
  dau_status_changed = TRUE;
}


void
pginc_proc(void) {
  get_page(PRI_FGROUND, inc_vtxpage(vtxwin.pgnum), 0, 0, vtx_buffer, &vtxwin);
}


void
pgdec_proc(void) {
  get_page(PRI_FGROUND, dec_vtxpage(vtxwin.pgnum), 0, 0, vtx_buffer, &vtxwin);
}


void
subpginc_proc(void) {
  int nextsubpage, subpages;
  
  nextsubpage = inc_vtxpage(vtxwin.subpgnum + 0x100) - 0x100;
  subpages = toptext_numsubpg(vtxwin.pgnum);
  if (subpages == 1) {
    nextsubpage = 0;
  } else if (subpages > 1 && nextsubpage > subpages) {
    nextsubpage = 1;
  }
  if (nextsubpage > 0x79) {
    nextsubpage = 1;
  }
  get_page(PRI_FGROUND_NH, vtxwin.pgnum, nextsubpage, 0, vtx_buffer, &vtxwin);
}


void
subpgdec_proc(void) {
  int prevsubpage, subpages;
  
  prevsubpage = dec_vtxpage(vtxwin.subpgnum + 0x200) - 0x200;
  subpages = toptext_numsubpg(vtxwin.pgnum);
  if (subpages == 1) {
    prevsubpage = 0;
  } else if (prevsubpage < 1) {
    if (subpages > 1) {
      prevsubpage = subpages;
    } else {
      prevsubpage = 0x79;
    }
  }
  get_page(PRI_FGROUND_NH, vtxwin.pgnum, prevsubpage, 0, vtx_buffer, &vtxwin);
}


void
subpgroll_proc(void) {
  get_page(PRI_FGROUND_NH, vtxwin.pgnum, 0, 0, vtx_buffer, &vtxwin);
}


static void
check_get_pgnum(int priority) {
  char *page_str, *subpage_str;
  int page, subpage;

  page_str = (char*)xv_get(pgnum_item, PANEL_VALUE);
  page = (strcmp(page_str, "") ? strtol(page_str, NULL, 16) : vtxwin.pgnum);
  subpage_str = (char*)xv_get(subpgnum_item, PANEL_VALUE);
  subpage = strtol(subpage_str, NULL, 16);

  if (vtx_chkpgnum(page, TRUE) && subpage >= 0 && subpage <= 0x7f) {
    get_page(priority, page, subpage, 0, vtx_buffer, &vtxwin);
    xv_set(pgnum_item, PANEL_VALUE, "", NULL);
    xv_set(subpgnum_item, PANEL_VALUE, "", NULL);
  } else {
    confirm_notice(frame, TRUE,
        "Invalid page number. Valid range\nis 100-8FF/1-7F (hexadecimal).");
  }
}


Panel_setting
pgnum_proc(void) {
  check_get_pgnum(PRI_FGROUND);
  return PANEL_NONE;
}


Panel_setting
pgnum_bg_proc(void) {
  check_get_pgnum(PRI_HIGH);
  return PANEL_NONE;
}


static void
display_tv(int x1, int y1, int x2, int y2, int hidden, const byte_t *buffer,
    const vtx_pageinfo_t *info) {
  byte_t tmp_buf[VTX_PAGESIZE];
  int pos;

  if (!vtx_dev_open || !config_get_tv_disp())
    return;
  if (hidden) {
    for (pos = x1 + y1 * 40; pos <= x2 + y2 * 40; pos++) {
      tmp_buf[pos - x1 - y1 * 40] = buffer[pos] & 0x7f;
    }
  } else {
    for (pos = x1 + y1 * 40; pos <= x2 + y2 * 40; pos++) {
      tmp_buf[pos - x1 - y1 * 40] = (((buffer[pos] & 0x7f) == 0x18) ? ' ' : buffer[pos] & 0x7f);
    }
  }
  CCTCHK(cct_putpage(x1, y1, x2, y2, tmp_buf, info), "cct_putpage", return);
}


void
update_pagenumdisp(int page, int subpage, int numsubpages, int cachecount) {
  char tmpstr[64];

  if (page < 0 || subpage < 0) {
    strcpy(tmpstr, "?/");
  } else {
    sprintf(tmpstr, "%X", page);
    if (subpage) {
      sprintf(tmpstr + strlen(tmpstr), ".%X", subpage);
    }
    strcat(tmpstr, "/");
  }
  if (numsubpages < 0) {
    strcat(tmpstr, "?/");
  } else {
    sprintf(tmpstr + strlen(tmpstr), "%d/", numsubpages);
  }
  if (page < 0 || subpage < 0) {
    strcat(tmpstr, "?");
  } else {
    sprintf(tmpstr + strlen(tmpstr), "%d", cachecount);
  }
  xv_set(pgmsg_item, PANEL_LABEL_STRING, tmpstr, NULL);
}


/* This gets called whenever the currently displayed page-number *may* have changed */
static void
pagenum_notify(void) {
  fcs_update_name();
}


/* Search for a DAU that is searching for a page. If no DAU is running, start the last one and
 * search for nonexistent page. As soon as another DAU starts searching, stop the one we started
 * and use the new one instead.
 */
static int
find_running_dau(void) {
  static int last_dau;
  int dau;

  /* No background DAUs available, use foreground-DAU */
  if (vtx_info.numpages == 1) {
#ifdef DEBUG_CHECK_STATION
    fprintf(stderr, "find_running_dau: no background DAUs, returning 0\n");
#endif
    return 0;
  /* Check whether the last DAU used is still searching & find another one if not */
  } else if (!last_dau || !dau_status[last_dau].page) {
    for (dau = 1; dau < vtx_info.numpages; dau++) {
      /* We found a running DAU */
      if (dau_status[dau].page) {
        /* Stop the last DAU we started (if any) */
        if (last_dau && !dau_status[last_dau].page) {
#ifdef DEBUG_CHECK_STATION
          fprintf(stderr, "find_running_dau: stopping DAU %d\n", last_dau);
#endif
          CCTCHK(cct_stop_dau(last_dau), "cct_stop_dau", return -1);
        }
#ifdef DEBUG_CHECK_STATION
        fprintf(stderr, "find_running_dau: using DAU %d\n", dau);
#endif
        return last_dau = dau;
      }
    }
    /* No DAU is searching for a page. Start the last one and search for a page that is unlikely
     * to exist to get the current running header-line.
     */
    last_dau = dau - 1;
#ifdef DEBUG_CHECK_STATION
    fprintf(stderr, "find_running_dau: starting DAU %d\n", last_dau);
#endif
    CCTCHK(cct_searchpage(0x8ff, 0x3f, 0x7f, PGMASK_PAGE | PGMASK_HOUR | PGMASK_MINUTE,
        last_dau), "cct_searchpage", return -1);
    CCTCHK(cct_reset_pgfound(last_dau), "cct_reset_pgfound", return -1);
    return last_dau;
  } else {
#ifdef DEBUG_CHECK_STATION
    fprintf(stderr, "find_running_dau: DAU %d still running\n", last_dau);
#endif
    return last_dau;
  }
}


#define STL_NO_HINT 0
#define STL_CHK_DAU0 3
#define STL_COUNT 5

static int
check_station_name(void) {
  static int last_dau;
  
  if (!vtx_dev_open)
    return 0;

  /* Try to find out the current station if it is unknown or if config_auto_reset()
   * returns TRUE.
   */
  if (!hotlist_get_current() || (config_get_auto_reset() && !station_selected)) {
    int dau, station, hint;
    static char station_name[STATIONNAME_MAXLEN], last_station_name[STATIONNAME_MAXLEN];
    static int station_loop, last_recognized, stl_check_dau0 = TRUE;

#ifdef DEBUG_CHECK_STATION
    fprintf(stderr, "check_station_name: current station: %d\n", hotlist_get_current());
#endif

    /* Get number of a running DAU */
    if ((dau = find_running_dau()) < 0)
      return -1;
    last_dau = dau;

    /* State counter for special actions */
    if (++station_loop >= STL_COUNT) {
      station_loop = 0;
      stl_check_dau0 = TRUE;
    }
#ifdef DEBUG_CHECK_STATION
    fprintf(stderr, "check_station_name: station_loop: %d\n", station_loop);
#endif

    /* If the header in the background-DAUs didn't change in cycle 0-2 (i.e. there's no running
     * page-number), we try reading the header from DAU 0. This is necessary because some
     * stations don't broadcast a header while the DAU is searching for a page, so the last
     * received one will stay there until a page is found (which might never happen with the
     * current driver-implementation, because for the SAA5243/6, we use an ugly kludge and
     * always read the header from buffer 4). Don't ask me why and how they do this, but in DAU 0
     * we should find a newer station-name if a page was already found there. If no page
     * was found, we're out of luck, but this method is better than not recognizing the station
     * at all.
     */
    if (stl_check_dau0 &&
        (station_loop == STL_CHK_DAU0 ||
        (station_loop == STL_CHK_DAU0 + 1 && last_recognized))) {
      dau = 0;
    }
    CCTCHK(cct_getpage(dau, 8, 0, 31, 0, station_name, NULL), "cct_getpage", return -1);
#ifdef DEBUG_CHECK_STATION
    fprintf(stderr, "check_station_name: read header from DAU %d\n", dau);
#endif

    /* Check whether the header has changed since the last iteration and don't check DAU 0 in
     * this cycle if it has. We can skip the check if stl_check_dau0 was already reset in this
     * cycle and we have to skip it, if the last header was read from DAU 0 (in that case, we
     * have to make sure the the next header also gets read from DAU 0 to get meaningful
     * results, which can only happen if stl_check_dau0 is still set).
     */
    if (stl_check_dau0 && station_loop != STL_CHK_DAU0 &&
        memcmp(station_name, last_station_name, sizeof(station_name))) {
      stl_check_dau0 = FALSE;
#ifdef DEBUG_CHECK_STATION
      fprintf(stderr, "check_station_name: header changed, won't check DAU 0\n");
#endif
    }
    memcpy(last_station_name, station_name, sizeof(station_name));

    /* Normally, we pass the current station as a hint to hotlist_search(), because a full search
     * is rather expensive. Every 5 seconds we do a full search nevertheless, because using hints
     * may fail if the station changed but the last station's name is a valid substring of the
     * current station's name.
     */
    hint = last_recognized;
    if (station_loop == STL_NO_HINT)
      hint = 0;
#ifdef DEBUG_CHECK_STATION
    fprintf(stderr, "check_station_name: using hint %d\n", hint);
#endif

    /* Try to recognize the current station. If the recognition is successful and the recognized
     * station is the same as the current station, we won't check DAU 0 in this cycle. If we
     * recognize a new station for the 2nd time, we set it as the current station.
     */
    if ((station = hotlist_search(station_name, hint))) {
#ifdef DEBUG_CHECK_STATION
      fprintf(stderr, "check_station_name: hotlist_search returned %d\n", station);
#endif
      if (station != hotlist_get_current() && station == last_recognized) {
        if (new_station(station, hotlist_get_current()) < 0)
          return -1;
#ifdef DEBUG_CHECK_STATION
        fprintf(stderr, "check_station_name: set new station %d\n", station);
#endif
      }
    }
    last_recognized = station;
#ifdef DEBUG_CHECK_STATION
    fprintf(stderr, "\n");
#endif

  } else {
    /* No station-name required. Stop DAU if it is still running and if it was started by us. */
    if (last_dau && !dau_status[last_dau].page) {
#ifdef DEBUG_CHECK_STATION
      fprintf(stderr, "check_station_name: stopping DAU %d\n\n", last_dau);
#endif
      CCTCHK(cct_stop_dau(last_dau), "cct_stop_dau", return -1);
      last_dau = 0;
    }
  }
  return 0;
}


Notify_value
update_proc(void) {
  int do_update = 0, st_changed, curr_st;
  static int flash_count = 0, lastfont = -1, last_dev_status = -1, check_name_loop;
  byte_t tmp_buffer[VTX_PAGESIZE];
  static byte_t stop_buffer[VTX_PAGESIZE];
  vtx_pageinfo_t tmp_info;
  static vtx_pageinfo_t stop_info;
  
  /* Load & display page in INtv-mode. This will be executed on startup and whenever VideoteXt
   * receives SIGUSR1
   */
  if (intv_fname && reload_intv_file) {
    FILE *file;
    int status;
    
    if (!((file = fopen(intv_fname, "r")))) {
      confirm_notice_v(frame, TRUE, "Can't open ", intv_fname, " for reading:\n", strerror(errno),
          (char *)0);
      xquit();
      return NOTIFY_DONE;
    }
    switch ((status = load_vtx(file, vtx_buffer, &vtxwin.page->info, NULL))) {
      case LOADERR:
        confirm_notice_v(frame, TRUE, "Error while reading ", intv_fname, ":\n", strerror(errno),
            (char *)0);
      break;
      case LOADEMAGIC:
        confirm_notice_v(frame, TRUE, "Error while reading ", intv_fname, ":\n"
            "Magic number missing. This is no VideoteXt- or INtv-file.", (char *)0);
      break;
      case LOADEVERSION:
        confirm_notice_v(frame, TRUE, "Error while reading ", intv_fname, ":\n"
            "Incompatible file format, possibly written by old version.", (char *)0);
      break;
      case LOADECORRUPT:
        confirm_notice_v(frame, TRUE, "Error while reading ", intv_fname, ":\n"
            "File is corrupt.", (char *)0);
      break;
    }
    fclose(file);
    if (status < 0) {
      xquit();
      return NOTIFY_DONE;
    }
    selection_hide(&vtxwin);
    decode_page(vtx_buffer, vtxwin.page, 0, 23);
    xv_set(checkbox, PANEL_VALUE, (int)xv_get(checkbox, PANEL_VALUE) | 1, NULL);
    vtxwin.stopped = TRUE;
    vtxwin.update = SCR_UPDATE;
    update_pagenumdisp(vtxwin.page->info.pagenum, vtxwin.page->info.minute, 1, 1);
    reload_intv_file = FALSE;
  }

  /* Check if a new page was received in DAU 0 & prepare to update the screen if necessary
   */
  if (!vtx_dev_open || (update_count && !dau_status[0].page_found))
    goto read_page_end;
    
  if (dau_status[0].page_found) {
    dau_status[0].page_found = FALSE;
    CCTCHK(cct_getpage(0, 0, 0, 39, 23, tmp_buffer, &tmp_info), "cct_getpage",
        return NOTIFY_DONE);

    /* Check if we received the right page (normally, the following test should never be
     * true, but you'll never know what these broken videotext-chips come up with next...)
     */
    if (dau_status[0].page != tmp_info.pagenum) {
      /* Restart DAU so that the header-line keeps on running */
      if (dau_status[0].keep_running) {
        CCTCHK(cct_searchpage(dau_status[0].page, 0, 0, PGMASK_PAGE, 0), "cct_searchpage",
            return NOTIFY_DONE);
        CCTCHK(cct_reset_pgfound(0), "cct_reset_pgfound", return NOTIFY_DONE);
      }
      goto read_page_end;
    }
    
    /* Sanity check: Test if page-header matches recognized station before inserting
     * page into cache
     */
    curr_st = hotlist_get_current();
    if (hotlist_search(tmp_buffer + 8, curr_st) == curr_st) {
      cache_insert(tmp_buffer, &tmp_info);
      search_bg_newpage(tmp_buffer, &tmp_info, search_notify_cb);
    }
    timeout_set(dau_status[0].page, FALSE);

    if (!hotlist_get_topdisable()) {
      int toptext_last = toptext_ok;
      
      toptext_newpage(tmp_buffer + 40, &tmp_info, &top_getpage);
      if (toptext_ok != toptext_last) {
        toptext_ok_notify();
      }
    }
    
    if (!hotlist_get_fastdisable()) {
      int fastext_last = fastext_ok;
      byte_t packet24[40], packet27[4 * 40];
      
      CCTCHK(cct_getpage(0, 0, 44, 39, 44, packet24, NULL), "cct_getpage",
          return NOTIFY_DONE);
      CCTCHK(cct_getpage(0, 0, 40, 39, 43, packet27, NULL), "cct_getpage",
          return NOTIFY_DONE);
      fastext_newpage(packet24, packet27, &tmp_info);
      if (fastext_ok != fastext_last) {
        fastext_ok_notify();
      }
      fastext_new_buttons(tmp_info.pagenum, tmp_info.minute);
    }
    
    if ((int)xv_get(cont_button, PANEL_CLIENT_DATA)) {
      xv_set(cont_button,
          PANEL_VALUE, TRUE,
          PANEL_CLIENT_DATA, FALSE,
          NULL);
    }

    if (dau_status[0].lastsubpg >= 0 && dau_status[0].subpage >= 0 && tmp_info.minute &&
        dau_status[0].lastsubpg != tmp_info.minute) {
      dau_status[0].lastsubpg = tmp_info.minute;
      if (!dau_status[0].firstsubpg) {
        dau_status[0].firstsubpg = tmp_info.minute;
      } else if (tmp_info.minute == dau_status[0].firstsubpg ||
          tmp_info.minute == dau_status[0].firstsubpg - 1) {
        /* All subpages are complete */
        dau_status[0].lastsubpg = -1;
        xv_set(cont_button, PANEL_VALUE, FALSE, NULL);
        cache_mark_complete(tmp_info.pagenum);
      }
    }
    if ((int)xv_get(cont_button, PANEL_VALUE)) {
      if (dau_status[0].subpage < 0 || dau_status[0].subpage == tmp_info.minute ||
          !tmp_info.minute) {
        xv_set(cont_button, PANEL_VALUE, FALSE, NULL);
      }
    }
    if (!tmp_info.minute) {
      cache_mark_complete(tmp_info.pagenum);
    }

    /* Update screen only, if the right subpage was found (or if we didn't search for a certain
     * subpage
     */
    if (dau_status[0].subpage <= 0 || dau_status[0].subpage == tmp_info.minute) {
      if (vtxwin.stopped || vtxwin.sel.show) {
        memcpy(stop_buffer + 40, tmp_buffer + 40, VTX_PAGESIZE - 40);
        stop_info = tmp_info;
        CCTCHK(cct_getpage(0, 7, 0, 39, 0, stop_buffer + 7, NULL), "cct_getpage",
            return NOTIFY_DONE);
        history_set_status(dau_status[0].page, STAT_FOUND);
      } else {
        stop_info.pagenum = 0;
        memcpy(vtx_buffer + 40, tmp_buffer + 40, VTX_PAGESIZE - 40);
        page.info = tmp_info;
        CCTCHK(cct_getpage(0, 7, 0, 39, 0, vtx_buffer + 7, NULL), "cct_getpage",
            return NOTIFY_DONE);
        vtx_buffer[0] = vtx_mkparity(7);
        decode_page(vtx_buffer, &page, 0, 23);
        history_set_status(dau_status[0].page, STAT_READ);
        update_pagenumdisp(tmp_info.pagenum, tmp_info.minute, toptext_numsubpg(tmp_info.pagenum),
            cache_count_subpg(tmp_info.pagenum));
        do_update = SCR_UPDATE;
      }
      CCTCHK(cct_reset_pgfound(0), "cct_reset_pgfound", return NOTIFY_DONE);
      dau_status[0].keep_running = FALSE;
    } else {
      history_set_status(dau_status[0].page, STAT_FOUND);
      /* Update the cached:-line if necessary */
      if (vtxwin.page->info.pagenum == tmp_info.pagenum) {
        update_pagenumdisp(tmp_info.pagenum, vtxwin.page->info.minute,
            toptext_numsubpg(tmp_info.pagenum),
            cache_count_subpg(tmp_info.pagenum));
      }
      /* Restart DAU so that the header-line keeps on running */
      if (dau_status[0].keep_running) {
        CCTCHK(cct_searchpage(dau_status[0].page, 0, 0, PGMASK_PAGE, 0), "cct_searchpage",
            return NOTIFY_DONE);
        CCTCHK(cct_reset_pgfound(0), "cct_reset_pgfound", return NOTIFY_DONE);
      }
    }
  } else { /* !page_found */
    int pagestat = 0;
  
    do {
      if (pagestat < 0 && report_cct_error(pagestat, "cct_checkpage")) {
        cct_close();
        hotlist_set_current(0);
        vtx_dev_open = FALSE;
        return -1;
      }
    } while ((pagestat = cct_checkpage(0)) < 0);
    if (!pagestat) {
      dau_status[0].page_found = TRUE;
    }
    
    /* If there's still something in the stop buffer, the display isn't stopped and we're still
     * searching for the same page, copy the contents from the stop buffer now.
     */
    if (!vtxwin.stopped && !vtxwin.sel.show && dau_status[0].page == stop_info.pagenum
        && (dau_status[0].subpage <= 0 || dau_status[0].subpage == stop_info.minute)) {
      memcpy(vtx_buffer + 7, stop_buffer + 7, VTX_PAGESIZE - 7);
      page.info = stop_info;
      vtx_buffer[0] = vtx_mkparity(7);
      decode_page(vtx_buffer, &page, 0, 23);
      history_set_status(dau_status[0].page, STAT_READ);
      update_pagenumdisp(stop_info.pagenum, stop_info.minute, toptext_numsubpg(stop_info.pagenum),
          cache_count_subpg(stop_info.pagenum));
      do_update = SCR_UPDATE;
      stop_info.pagenum = 0;
    }
  }
  
  /* If it's time to update the screen, check if a new page was received. If there's a new
   * page, delay the update, because we don't know yet if it really was the page we were
   * searching for, so we don't want to stop the running header-line now.
   */
  if (!do_update && !vtxwin.stopped && !vtxwin.sel.show && config_get_update_speed()
      && !dau_status[0].page_found) {
    CCTCHK(cct_getpage(0, 7, 0, 39, 0, vtx_buffer + 7, NULL), "cct_getpage", return NOTIFY_DONE);
    decode_page(vtx_buffer, &page, 0, 0);
    do_update = HDR_UPDATE;
  }

read_page_end:
  /* Update the screen if one of the update-flags is set or if it's time to enter a new flash-cycle
   */
  if (--update_count < 0) {
    update_count = update_table[config_get_update_speed()].update_interval;
  }
  switch (MAX(do_update, vtxwin.update)) {
    case HDR_UPDATE:
      x_vtx_redraw(&vtxwin, 0, 0, 39, 0, FALSE);
      display_tv(0, 0, 39, 0, vtxwin.hidden, vtx_buffer, &page.info);
    break;
    case SCR_UPDATE:
      x_vtx_redraw(&vtxwin, 0, 0, 39, 23, FALSE);
      display_tv(0, 0, 39, 23, vtxwin.hidden, vtx_buffer, &page.info);
      pagenum_notify();
    break;
  }
  if (!flash_count--) {
    if (config_get_blinkdis()) {
      flash_count = 6;
      if (vtxwin.flash_state) {
        vtxwin.flash_state = 0;
        x_vtx_redraw(&vtxwin, 0, 0, 39, 23, FALSE);
      }
      vtxwin.flash_state = 0;
    } else {
      flash_count = (vtxwin.flash_state = !vtxwin.flash_state) ? 3 : 6;
      x_vtx_redraw(&vtxwin, 0, 0, 39, 23, FALSE);
    }
  }
  vtxwin.update = 0;

  if (queue_itimer() < 0)
    return NOTIFY_DONE;
  if (top_flush_getpage() < 0) {
    return NOTIFY_DONE;
  }

  /* Display cct-status on left footer
   */
  if (dau_status_changed || vtx_dev_open != last_dev_status) {
    int pgbuf;
    char tmpstr[256] = "CCT Status:", *strend;
    
    if (intv_fname) {
      xv_set(frame,
          FRAME_LEFT_FOOTER, "Display-only mode",
          FRAME_RIGHT_FOOTER, intv_fname,
          NULL);
    } else {
      strend = tmpstr + strlen(tmpstr);
      if (vtx_dev_open) {
        for (pgbuf = 0; pgbuf < vtx_info.numpages; pgbuf++) {
          strend += sprintf(strend, " %d:", pgbuf);
          if (dau_status[pgbuf].page) {
            strend += sprintf(strend, "%X", dau_status[pgbuf].page);
          } else {
            strend += sprintf(strend, "Idle");
          }
          if (strend - tmpstr > 200)
            break;
        }
      } else {
        sprintf(strend, " Not connected%s", keep_closed ? " (locked)" : "");
      }
      xv_set(frame, FRAME_LEFT_FOOTER, tmpstr, NULL);
    }
    dau_status_changed = FALSE;
  }

  /* Check the current station once a second */
  if (++check_name_loop > CHECK_NAME_CYCLE) {
    check_name_loop = 0;
    if (check_station_name() < 0)
      return NOTIFY_DONE;
  }

  /* Display station-name & charset-info on right footer if station has changed */
  if (!intv_fname &&
      ((st_changed = hotlist_station_changed()) || vtxwin.page->info.charset != lastfont)) {
    char *tmp_name;
    
    tmp_name = sstrdup(hotlist_get_name());
    if (vtxwin.page->info.charset == 3 || vtxwin.page->info.charset == 7) {
      tmp_name = sstrapp(tmp_name, "  (Unknown charset)");
    }
    xv_set(frame, FRAME_RIGHT_FOOTER, tmp_name, NULL);
    free(tmp_name);
    if (st_changed) {
      sel_st_update(TRUE);
      hot_update();
      station_opt_update();
      spool_opt_update();
      tuning_update(TRUE);
    }
  }
  lastfont = vtxwin.page->info.charset;
  last_dev_status = vtx_dev_open;
  return NOTIFY_DONE;
}


void
canvas_repaint(Canvas canvas, Xv_window paint_window, const Rectlist *area) {
  x_vtx_redraw(&vtxwin, area->rl_bound.r_left / vtxfonts[vtxwin.font].width,
      area->rl_bound.r_top / vtxfonts[vtxwin.font].height,
      (area->rl_bound.r_width  + area->rl_bound.r_left) / vtxfonts[vtxwin.font].width,
      (area->rl_bound.r_height + area->rl_bound.r_top) / vtxfonts[vtxwin.font].height, TRUE);
}


void
canvas_proc(Xv_window window, const Event *event) {
  int pgnum, action;
  static int click_x, click_y;

  switch (action = event_action(event)) {
    case ACTION_SELECT:
    case ACTION_MENU:
      if (event_is_down(event)) {
        click_x = event_x(event);
        click_y = event_y(event);
        selection_finish(&vtxwin);
        selection_start(&vtxwin, click_x / vtxfonts[vtxwin.font].width,
            click_y / vtxfonts[vtxwin.font].height, event_ctrl_is_down(event),
            (action == ACTION_MENU));
      } else if (event_is_up(event)) {
        int drag = vtxwin.sel.drag;
        
        selection_finish(&vtxwin);
        if (!intv_fname && !drag && (pgnum = x_extract_pgnum(&vtxwin,
            click_x / vtxfonts[vtxwin.font].width, click_y / vtxfonts[vtxwin.font].height))) {
          get_page(action = ACTION_SELECT ? PRI_FGROUND : PRI_HIGH, pgnum, 0, 0, vtx_buffer,
              &vtxwin);
        }
      }
    break;
    case LOC_DRAG:
      if ((action_select_is_down(event) || action_menu_is_down(event))
          && (vtxwin.sel.drag || abs(click_x - event_x(event)) > DRAG_THRES
          || abs(click_y - event_y(event)) > DRAG_THRES)) {
        selection_drag(&vtxwin, event_x(event) / vtxfonts[vtxwin.font].width,
            event_y(event) / vtxfonts[vtxwin.font].height);
      }
    break;
    case KBD_USE:
      xv_set(right_panel, WIN_SET_FOCUS, NULL);
    break;
  }
}


void
frame_proc(Xv_window window, const Event *event) {
  switch (event_action(event)) {
    case WIN_RESIZE: {
      int font, foundfont = FALSE;

      xv_set((Menu_item)xv_get(font_menu, MENU_NTH_ITEM, vtxwin.font + 1),
          MENU_SELECTED, FALSE, NULL);

      vtxwin.font = 0;		/* Find the biggest font that fits into the canvas */
      for (font = 0; font <= vtxmaxfont; font++) {
        if (vtxfonts[vtxwin.font].width <= vtxfonts[font].width &&
            vtxfonts[vtxwin.font].height <= vtxfonts[font].height &&
            vtxfonts[font].width * 41 <= (int)xv_get(frame, XV_WIDTH) -
            (intv_fname ? 0 : RPANEL_WIDTH) - 4 * WIN_DEFAULT_BORDER_WIDTH &&
            vtxfonts[font].height * 26 <= (int)xv_get(frame, XV_HEIGHT) - CANVAS_OFFSET) {
          vtxwin.font = font;
          foundfont = TRUE;
        }
      }
      if (!foundfont) {		/* If no font fits, find the smallest one */
        for (font = 0; font <= vtxmaxfont; font++) {
          if (vtxfonts[font].width <= vtxfonts[vtxwin.font].width &&
              vtxfonts[font].height <= vtxfonts[vtxwin.font].height)
            vtxwin.font = font;
        }
      }

      xv_set((Menu_item)xv_get(font_menu, MENU_NTH_ITEM, vtxwin.font + 1),
          MENU_SELECTED, TRUE, NULL);
      xv_set(vtxwin.canvas_bg, /*FIXME: Don't use globals */
          XV_WIDTH, 41 * vtxfonts[vtxwin.font].width + 4 * WIN_DEFAULT_BORDER_WIDTH,
          XV_HEIGHT, 26 * vtxfonts[vtxwin.font].height,
          NULL);
      xv_set(vtxwin.canvas, /*FIXME: Don't use globals */
          XV_Y, vtxfonts[vtxwin.font].height,
          XV_WIDTH, 40 * vtxfonts[vtxwin.font].width + 2 * WIN_DEFAULT_BORDER_WIDTH,
          XV_HEIGHT, 24 * vtxfonts[vtxwin.font].height + 2 * WIN_DEFAULT_BORDER_WIDTH,
          CANVAS_WIDTH, 40 * vtxfonts[vtxwin.font].width + 2 * WIN_DEFAULT_BORDER_WIDTH,
          CANVAS_HEIGHT, 24 * vtxfonts[vtxwin.font].height + 2 * WIN_DEFAULT_BORDER_WIDTH,
          NULL);
      /* Need to set XV_WIDTH to work around bug in XView (otherwise parts exposed after moving
       * don't get redrawn)
       */
      xv_set(right_panel,
          XV_X, (int)xv_get(frame, XV_WIDTH) - RPANEL_WIDTH,
          XV_WIDTH, RPANEL_WIDTH,
          NULL);
      xv_set(intv_reveal_toggle, XV_X, (int)xv_get(frame, XV_WIDTH) - 65, NULL);
      position_popups();
    }
    break;
    case WIN_REPARENT_NOTIFY: {
      static int popup_init;

      /* Creating the popups here makes sure that we can find out the size of the wm's decoration
       * since our frame-window has already been reparented by the wm. Unfortunately, this doesn't
       * work if the window manager doesn't reparent our window (or if no wm is running at all),
       * but since the popups are not vital, I ignore this.
       */
      if (!popup_init && !intv_fname) {
        popup_init = TRUE;
        if (config_get_startup(POPUP_SEARCH))
          search_proc();
        if (config_get_startup(POPUP_HISTORY))
          history_proc();
        if (config_get_startup(POPUP_TOPTEXT))
          toptext_proc();
        if (config_get_startup(POPUP_HOTLIST))
          hotlist_proc();
        if (config_get_startup(POPUP_SELST))
          sel_station_proc();
      }
    }
    break;
    case WIN_PROPERTY_NOTIFY: {
      static int laststate = -1;
      
      if ((int)xv_get(frame, FRAME_CLOSED) != laststate) {
        laststate = (int)xv_get(frame, FRAME_CLOSED);
        if (config_get_group_iconify()) {
          iconify_popups(laststate);
        }
      }
    }
    break;
  }
}


static Notify_value
check_hotkeys(Panel panel, const Event *event, Notify_arg arg, Notify_event_type type,
    int need_meta) {
  XKeyEvent *xev = (XKeyEvent*)event->ie_xevent;
  KeySym keysym;
  char buf[3];

  /* XView sometimes sends us events without a XEvent-structure, so we better check for NULL
   * pointers here. Of course, nobody tells you that you have to do it. :-(
   */
  if (xev && xev->type == KeyPress) {
    buf[XLookupString(xev, buf, sizeof buf, &keysym, &compose_status)] = '\0';
    if (strlen(buf) == 1 &&
        ((!isxdigit(buf[0]) && !iscntrl(buf[0]) && !need_meta) || xev->state & Mod1Mask)) {
      switch (tolower(buf[0])) {
        case 'e':
          xv_set(checkbox, PANEL_VALUE, (int)xv_get(checkbox, PANEL_VALUE) ^ 2, NULL);
          checkbox_proc();
          return NOTIFY_DONE;
        case 'x':
          if (xev->state & Mod1Mask) {
            xquit();
            return NOTIFY_DONE;
          }
        break;
      }
      if (!intv_fname) {
        switch (tolower(buf[0])) {
          case '\r':
          case '\n':
            pgnum_bg_proc();
            return NOTIFY_DONE;
          case '+':
            pginc_proc();
            return NOTIFY_DONE;
          case '-':
            pgdec_proc();
            return NOTIFY_DONE;
          case 's':
            xv_set(checkbox, PANEL_VALUE, (int)xv_get(checkbox, PANEL_VALUE) ^ 1, NULL);
            checkbox_proc();
            return NOTIFY_DONE;
          case 'r':
            if (xev->state & Mod1Mask) {
              reset_proc();
              return NOTIFY_DONE;
            }
          break;
          case 'n':
            history_move(+1);
            return NOTIFY_DONE;
          case 'p':
            history_move(-1);
            return NOTIFY_DONE;
          case 'b':
            toptext_go_nextblk();
            return NOTIFY_DONE;
          case 'g':
            toptext_go_nextgrp();
            return NOTIFY_DONE;
          case 'i':
            get_page(PRI_FGROUND, hotlist_get_index_page(), 0, 0, vtx_buffer, &vtxwin);
            return NOTIFY_DONE;
          case 'l':
            subpgroll_proc();
            return NOTIFY_DONE;
          case 'o':
            xv_set(cont_button, PANEL_VALUE, !(int)xv_get(cont_button, PANEL_VALUE), NULL);
            return NOTIFY_DONE;
        }
      }
    }
    if (!intv_fname) {
      switch (keysym) {
        case XK_Up:
          subpginc_proc();
          return NOTIFY_DONE;
        case XK_Down:
          subpgdec_proc();
          return NOTIFY_DONE;
        case XK_Prior:
          get_next_toppage(TRUE);
          return NOTIFY_DONE;
        case XK_Next:
          get_prev_toppage(TRUE);
          return NOTIFY_DONE;
      }
    }
  }
  return notify_next_event_func(panel, (Notify_event)event, arg, type);
}


/* These functions check if a hotkey was pressed. For check_hotkeys_nometa, most hotkeys can
 * also be pressed without Meta (except for 'x' and 'r'). Hex-digits must be pressed together
 * with Meta, otherwise they'll get confused with page-numbers.
 */
Notify_value
check_hotkeys_meta(Panel panel, const Event *event, Notify_arg arg, Notify_event_type type) {
  return check_hotkeys(panel, event, arg, type, TRUE);
}


Notify_value
check_hotkeys_nometa(Panel panel, const Event *event, Notify_arg arg, Notify_event_type type) {
  return check_hotkeys(panel, event, arg, type, FALSE);
}


static Notify_value
frame_destroy_func(Notify_client client, Destroy_status status) {
  if (status == DESTROY_CLEANUP) {
    if (!doabort && !intv_fname) {
      search_reset();
      hotlist_exit();
      write_cache();
    }
    return notify_next_destroy_func(client, status);
  } else {
    return NOTIFY_DONE;
  }
}


void
x_main_loop(void) {
  int pos;
  struct itimerval itimer;

  for (pos = 0; pos <= 959; pos++) {
    vtx_buffer[pos] = ' ';
    page.chr[pos] = titlepage_chr[pos];
    page.attrib[pos] = titlepage_attrib[pos];
    curr_disp.chr[pos] = ' ';
    curr_disp.attrib[pos] = 65535;
  }

  vtxwin.canvas_bg = (Canvas)xv_create(frame, CANVAS,
      XV_X, 0,
      XV_Y, CANVAS_OFFSET,
      CANVAS_AUTO_SHRINK, FALSE,
      CANVAS_AUTO_EXPAND, FALSE,
      WIN_CMS, cms,
      WIN_BACKGROUND_COLOR, CMSCBLACK,
      WIN_FOREGROUND_COLOR, CMSCWHITE,
      NULL);
  vtxwin.canvas = (Canvas)xv_create(vtxwin.canvas_bg, CANVAS,
      XV_X, 1,
      XV_Y, vtxfonts[0].height - 2 * WIN_DEFAULT_BORDER_WIDTH,
      CANVAS_AUTO_SHRINK, FALSE,
      CANVAS_AUTO_EXPAND, FALSE,
      CANVAS_REPAINT_PROC, canvas_repaint,
      CANVAS_X_PAINT_WINDOW, FALSE,
      CANVAS_FIXED_IMAGE, FALSE,
      WIN_CMS, cms,
      WIN_BACKGROUND_COLOR, CMSCBLACK,
      WIN_FOREGROUND_COLOR, CMSCBLACK,
      NULL);
  xv_set(canvas_paint_window(vtxwin.canvas),
      WIN_EVENT_PROC, canvas_proc,
      WIN_CONSUME_EVENTS,
        WIN_MOUSE_BUTTONS, KBD_USE, LOC_DRAG, NULL,
      NULL);
  vtxwin.winid = (Window)xv_get(canvas_paint_window(vtxwin.canvas), XV_XID);
  vtxwin.gc = vtxgc;
  vtxwin.flash_state = vtxwin.update = 0;
  vtxwin.page = &page;
  vtxwin.curr_disp = &curr_disp;
  vtxwin.pgnum = 0x100;
  vtxwin.subpgnum = 0;
  vtxwin.font = config_get_font();
  checkbox_proc();

  xv_set(font_menu, MENU_VALUE, vtxwin.font, NULL);
  xv_set(frame,
      XV_WIDTH, 41 * vtxfonts[vtxwin.font].width + (intv_fname ? 0 : RPANEL_WIDTH) +
          4 * WIN_DEFAULT_BORDER_WIDTH,
      XV_HEIGHT, 26 * vtxfonts[vtxwin.font].height + CANVAS_OFFSET,
      NULL);
  xv_set(right_panel, XV_X, 41 * vtxfonts[vtxwin.font].width + 4 * WIN_DEFAULT_BORDER_WIDTH, NULL);
  xv_set(intv_reveal_toggle, XV_X, 41 * vtxfonts[vtxwin.font].width + 2 * WIN_DEFAULT_BORDER_WIDTH
      - 65, NULL);
  xv_set(scr_menu, MENU_VALUE, config_get_update_speed(), NULL);
  hotlist_init();

  itimer.it_interval.tv_sec = 0; itimer.it_interval.tv_usec = 200000;
  itimer.it_value.tv_sec = 0; itimer.it_value.tv_usec = 200000;
  notify_set_itimer_func(frame, update_proc, ITIMER_REAL, &itimer, NULL);

  assert(notify_interpose_destroy_func(frame, frame_destroy_func) == NOTIFY_OK);

  xv_main_loop(frame);
}


/* Read 3-digit number at pos x,y in vtxwin, also do some *blinkenlights*
 * Return 0 if no number found, hex-value otherwise
 */
int
x_extract_pgnum(const vtxpgwin_t *vtxwin, int x, int y) {
  int numstart, numend, pos, pgnum;
  chr_t *chr, num[4];
  
  x = LIMIT(x, 0, 39);
  y = LIMIT(y, 0, 23);
  chr = vtxwin->page->chr;
  if (vtxwin->page->attrib[y * 40 + x] & VTX_HIDDEN && vtxwin->hidden)
    return 0;
  if (vtxwin->page->attrib[y * 40 + x] & VTX_DOUBLE1)
    y++;
  for (numstart = x; numstart >= 0; numstart--)
    if (!isdigit(chr[y * 40 + numstart]))
      break;
  numstart++;
  for (numend = x; numend <= 39; numend++)
    if (!isdigit(chr[y * 40 + numend]))
      break;
  numend--;
  if (numend - numstart != 2)
    return 0;
  memcpy(num, chr + y * 40 + numstart, 3);
  num[3] = 0;
  pgnum = strtol(num, NULL, 16);
  if (vtx_chkpgnum(pgnum, FALSE)) {
    for (pos = numstart + y * 40; pos <= numend + y * 40; pos++) {
      vtxwin->page->attrib[pos] ^= VTX_INVERT;
    }
    x_vtx_redraw(vtxwin, numstart, y, numend, y, FALSE);
    XSync(dpy, False);
    xv_sleep(0, 200000);
    for (pos = numstart + y * 40; pos <= numend + y * 40; pos++) {
      vtxwin->page->attrib[pos] ^= VTX_INVERT;
    }
    x_vtx_redraw(vtxwin, numstart, y, numend, y, FALSE);
    XSync(dpy, False);
    return pgnum;
  }
  return 0;
}


/* The next two routines are responsible for (re)drawing a VTX-page. flush_string_buffer()
 * should be called only by x_vtx_redraw() which tries to do as few GC-switches as possible and
 * also to print as long strings as possible to reduce X-protocol-overhead.
 * vtxwin is the structure that holds all necessary information about a VTX-page.
 * If expose is true, the complete requested region gets redrawn; otherwise only parts that have
 * changed since the last call get updated.
 * x1/y1 & x2/y2 (in character-positions, _not_ pixels!) define the rectangle that will be drawn
 * Returns TRUE if OK, FALSE if rectangle is invalid
 */
typedef struct {
  Window winid;
  GC gc;
  int xpos, ypos, count, xinc, yinc, flash_state;
  unsigned long fg, bg;
  attrib_t attr;
  chr_t *chr;
} stringbuf_t;

static void
flush_string_buffer(stringbuf_t *buf) {
  static chr_t tmp_str[40];
  int a;
     
  if (buf->count > 0) {
    if (buf->attr & VTX_HIDDEN || (buf->attr & VTX_FLASH && buf->flash_state)) {
      memset(tmp_str, ' ', buf->count);
    } else {
      memcpy(tmp_str, buf->chr, buf->count);
    }
    if (!do_dither) {
      XDrawImageString(dpy, buf->winid, buf->gc, buf->xpos, buf->ypos, tmp_str, buf->count);
    } else {
      XSetStipple(dpy, buf->gc, dith_bm[buf->bg]);
      XFillRectangle(dpy, buf->winid, buf->gc, buf->xpos, buf->ypos - buf->yinc,
          buf->xinc * buf->count, buf->yinc);
      XSetStipple(dpy, buf->gc, dith_bm[buf->fg]);
      XDrawString(dpy, buf->winid, buf->gc, buf->xpos, buf->ypos, tmp_str, buf->count);
    }
    if (buf->attr & VTX_GRSEP) {
      for (a = 0; a < buf->count; a++)
        tmp_str[a] = (buf->chr[a] >= 0x80 && buf->chr[a] <= 0xbf) ? 255 : ' ';
      if (!do_dither) {
        XSetForeground(dpy, buf->gc, buf->bg);
        XDrawString(dpy, buf->winid, buf->gc, buf->xpos, buf->ypos, tmp_str, buf->count);  
        XSetForeground(dpy, buf->gc, buf->fg);
      } else {
        XSetStipple(dpy, buf->gc, dith_bm[buf->bg]);
        XDrawString(dpy, buf->winid, buf->gc, buf->xpos, buf->ypos, tmp_str, buf->count);  
      }
    }
    buf->chr += buf->count;
    buf->xpos += buf->count * buf->xinc;
  }
  buf->count = 0;
}

int
x_vtx_redraw(const vtxpgwin_t *vtxwin, int x1, int y1, int x2, int y2, int expose) {
  int line, col, doubleht = 0;
  stringbuf_t buf;
  attrib_t *currattr, *currdispattr;
  chr_t *currchr, *currdispchr, tmp_chr;
  const attrib_t *lastattr, default_attr = 7;

  x1 = MAX(0, x1);
  y1 = MAX(0, y1);
  x2 = MIN(39, x2);
  y2 = MIN(23, y2);
  if (y1 * 40 + x1 > y2 * 40 + x2)
    return FALSE;
  buf.xinc = vtxfonts[vtxwin->font].width;
  buf.yinc = vtxfonts[vtxwin->font].height;
  buf.winid = vtxwin->winid;
  buf.attr = default_attr;
  buf.gc = vtxwin->gc;
  buf.flash_state = vtxwin->flash_state;
  lastattr = &default_attr;
  XSetForeground(dpy, vtxwin->gc, buf.fg = VTXCWHITE);
  XSetBackground(dpy, vtxwin->gc, buf.bg = VTXCBLACK);
  if (!do_dither) {
    XSetFillStyle(dpy, vtxwin->gc, FillSolid);
  } else {
    XSetFillStyle(dpy, vtxwin->gc, FillOpaqueStippled);
    buf.fg = 7;
    buf.bg = 0;
  }
  XSetFont(dpy, vtxwin->gc, vtxfonts[vtxwin->font].normal->fid);
  for (line = y1; (line <= y2 || doubleht == 2) && line <= 23; line++) {
    buf.count = 0;
    buf.xpos = x1 * buf.xinc;
    buf.ypos = (line + 1) * buf.yinc;
    buf.chr = currchr = vtxwin->page->chr + line * 40 + x1;
    currattr = vtxwin->page->attrib + line * 40 + x1;
    currdispchr = vtxwin->curr_disp->chr + line * 40 + x1;
    currdispattr = vtxwin->curr_disp->attrib + line * 40 + x1;
    if (doubleht) {
      doubleht--;
    }
    for (col = x1; col <= x2; col++) {
      if (*currattr & VTX_DOUBLE1) {
        doubleht = 2;
      }
      tmp_chr = ((*currattr & VTX_HIDDEN && vtxwin->hidden) ||
          (*currattr & VTX_FLASH && vtxwin->flash_state)) ? ' ' : *currchr;
      if (expose || *currdispchr != tmp_chr || *currdispattr != *currattr
          || ((*currattr & VTX_DOUBLE2) && doubleht)) {
        *currdispchr = tmp_chr;
        *currdispattr = *currattr;
        if ((*lastattr & VTX_COLMASK) != (*currattr & VTX_COLMASK)
            || (*lastattr & VTX_INVERT) != (*currattr & VTX_INVERT)) {
          flush_string_buffer(&buf);
          if (*currattr & VTX_INVERT) {
            buf.bg = *currattr & VTX_COLMASK;
          } else {
            buf.fg = *currattr & VTX_COLMASK;
          }
          if (!do_dither) {
            if (*currattr & VTX_INVERT) {
              buf.bg = vtxcolors[buf.bg];
              XSetBackground(dpy, vtxwin->gc, buf.bg);
            } else {
              buf.fg = vtxcolors[buf.fg];
              XSetForeground(dpy, vtxwin->gc, buf.fg);
            }
          }
        }
        if ((*lastattr & VTX_BGMASK) != (*currattr & VTX_BGMASK)
            || (*lastattr & VTX_INVERT) != (*currattr & VTX_INVERT)) {
          flush_string_buffer(&buf);
          if (*currattr & VTX_INVERT) {
            buf.fg = (*currattr >> 3) & VTX_COLMASK;
          } else {
            buf.bg = (*currattr >> 3) & VTX_COLMASK;
          }
          if (!do_dither) {
            if (*currattr & VTX_INVERT) {
              buf.fg = vtxcolors[buf.fg];
              XSetForeground(dpy, vtxwin->gc, buf.fg);
            } else {
              buf.bg = vtxcolors[buf.bg];
              XSetBackground(dpy, vtxwin->gc, buf.bg);
            }
          }
        }
        if ((*lastattr & VTX_DOUBLE2) != (*currattr & VTX_DOUBLE2)) {
          flush_string_buffer(&buf);
          XSetFont(dpy, vtxwin->gc, (*currattr & VTX_DOUBLE2 ?
              vtxfonts[vtxwin->font].doubleht->fid : vtxfonts[vtxwin->font].normal->fid));
        }
        if ((*lastattr & VTX_GRSEP) != (*currattr & VTX_GRSEP)) {
          flush_string_buffer(&buf);
          buf.attr = (buf.attr & ~VTX_GRSEP) | (*currattr & VTX_GRSEP);
        }
        if ((*lastattr & VTX_HIDDEN) != (*currattr & VTX_HIDDEN)) {
          flush_string_buffer(&buf);
          buf.attr = (buf.attr & ~VTX_HIDDEN) | (*currattr & VTX_HIDDEN) * vtxwin->hidden;
        }
        if ((*lastattr & VTX_FLASH) != (*currattr & VTX_FLASH)) {
          flush_string_buffer(&buf);
          buf.attr = (buf.attr & ~VTX_FLASH) | (*currattr & VTX_FLASH);
        }
        lastattr = currattr;
        buf.count++;
      } else {
        flush_string_buffer(&buf);
        buf.xpos += buf.xinc;
        buf.chr++;
      }
      currchr++;
      currattr++;
      currdispchr++;
      currdispattr++;
    }
    flush_string_buffer(&buf);
  }
  return TRUE;
}
